/*
 *  linux/arch/mips/kernel/proc.c
 *
 *  Copyright (C) 1995, 1996  Ralf Baechle
 */
#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <asm/bootinfo.h>
#include <asm/mipsregs.h>
#include <asm/processor.h>
#include <asm/watch.h>

unsigned long unaligned_instructions;
unsigned int vced_count, vcei_count;

#ifdef CONFIG_CPU_VR41XX

#include <asm/vr41xx.h>

inline int calc_pclock_from_clkspeedreg(unsigned short clkspeedreg)
{
	return ( (18432000 / (clkspeedreg & 0x1f)) * 64 );
}

int get_pclock(void)
{
	return calc_pclock_from_clkspeedreg(*VR41XX_CLKSPEEDREG);
}

int get_tclock(void)
{
	int ps_per_t;
	unsigned short clkspeedreg = *(VR41XX_CLKSPEEDREG);
	
	switch (*VR41XX_REVIDREG & 0xf000)
	{
		case 0x2000:
			// RID is for vr4111
			switch (clkspeedreg & 0xe000 )
			{
				case 0x6000:
					ps_per_t = 2;
					break;

				case 0xa000:
					ps_per_t = 3;
					break;

				case 0xc000:
					ps_per_t = 4;
					break;

				default:
					// unexpected DIVT
					return 0;
			}
			break;

		case 0x3000:
			// RID is for vr4121
			ps_per_t = (clkspeedreg & 0xf000) >> 12;
			break;

		default:
			// unexpected RID
			return 0;
	}

	return ( calc_pclock_from_clkspeedreg(clkspeedreg) / ps_per_t );
}
#endif

/*
 * BUFFER is PAGE_SIZE bytes long.
 *
 * Currently /proc/cpuinfo is being abused to print data about the
 * number of date/instruction cacheflushes.
 */
int get_cpuinfo(char *buffer)
{
	char fmt [64];
	const char *cpu_name[] = CPU_NAMES;
	const char *mach_group_names[] = GROUP_NAMES;
	const char *mach_unknown_names[] = GROUP_UNKNOWN_NAMES;
	const char *mach_jazz_names[] = GROUP_JAZZ_NAMES;
	const char *mach_dec_names[] = GROUP_DEC_NAMES;
	const char *mach_arc_names[] = GROUP_ARC_NAMES;
	const char *mach_sni_rm_names[] = GROUP_SNI_RM_NAMES;
	const char *mach_acn_names[] = GROUP_ACN_NAMES;
	const char *mach_sgi_names[] = GROUP_SGI_NAMES;
	const char *mach_cobalt_names[] = GROUP_COBALT_NAMES;
	const char *mach_vr41xx_names[] = GROUP_VR41XX_NAMES;
	const char *mach_poseidon_names[] = GROUP_POSEIDON_NAMES;
	const char **mach_group_to_name[] = {
		mach_unknown_names,
		mach_jazz_names,
		mach_dec_names,
		mach_arc_names,
		mach_sni_rm_names,
		mach_acn_names,
		mach_sgi_names,
		mach_cobalt_names,
		mach_vr41xx_names,
		mach_poseidon_names
	};

	unsigned int version = read_32bit_cp0_register(CP0_PRID);
	int len;

	len = sprintf(buffer, "cpu\t\t\t: MIPS\n");
	len += sprintf(buffer + len, "cpu model\t\t: %s V%d.%d\n",
	               cpu_name[mips_cputype <= CPU_LAST ?
	                        mips_cputype :
	                        CPU_UNKNOWN],
	               (version >> 4) & 0x0f,
	               version & 0x0f);
	len += sprintf(buffer + len, "system type\t\t: %s %s\n",
		       mach_group_names[mips_machgroup],
		       mach_group_to_name[mips_machgroup][mips_machtype]);

#ifdef CONFIG_CPU_VR41XX
	{
		int clock;
		clock = get_pclock();
		len += sprintf(buffer + len, "cpu_mhz\t\t\t: %d.%.6d\n", clock / 1000000, clock % 1000000);
		clock = get_tclock();
		len += sprintf(buffer + len, "bus_mhz\t\t\t: %d.%.6d\n", clock / 1000000, clock % 1000000);
	}
#endif

	len += sprintf(buffer + len, "BogoMIPS\t\t: %lu.%02lu\n",
		       (loops_per_sec + 2500) / 500000,
	               ((loops_per_sec + 2500) / 5000) % 100);
#if defined (__MIPSEB__)
	len += sprintf(buffer + len, "byteorder\t\t: big endian\n");
#endif
#if defined (__MIPSEL__)
	len += sprintf(buffer + len, "byteorder\t\t: little endian\n");
#endif
	len += sprintf(buffer + len, "unaligned accesses\t: %lu\n",
		       unaligned_instructions);
	len += sprintf(buffer + len, "wait instruction\t: %s\n",
	               wait_available ? "yes" : "no");
	len += sprintf(buffer + len, "microsecond timers\t: %s\n",
	               cyclecounter_available ? "yes" : "no");
	len += sprintf(buffer + len, "extra interrupt vector\t: %s\n",
	               dedicated_iv_available ? "yes" : "no");
	len += sprintf(buffer + len, "hardware watchpoint\t: %s\n",
	               watch_available ? "yes" : "no");

	sprintf(fmt, "VCE%%c exceptions\t\t: %s\n",
	        vce_available ? "%d" : "not available");
	len += sprintf(buffer + len, fmt, 'D', vced_count);
	len += sprintf(buffer + len, fmt, 'I', vcei_count);

	return len;
}
