/* $Id: e105btns.c,v 1.4 1999/11/15 05:34:45 mfklar Exp $
 *
 * arch/mips/vr41xx/e105btns.c
 *
 * Casio E-10x/50x input button driver
 *
 * Copyright (c) 1999, Robert Coie <rac@intrigue.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/init.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/vr41xx.h>

#define E105_BUTTONS_GPIO	9
#define E105_BUTTONS_IRQ	VR41XX_IRQ_GPIO(E105_BUTTONS_GPIO)
#define E105_BUTTONS_MASK	(1 << E105_BUTTONS_GPIO)
#define E105_BUTTONS		preg8(KSEG1 + 0x1400A018)
#define E105_BUTTONS_DEV_ID	((void *) 0x45357878)	/* 'E5xx' */

struct e105_buttons_device_t {
	__u8			status;
	unsigned int		ready;
	wait_queue_head_t	waitq;
	int			usage_count;
	struct fasync_struct *	fasync;
};

/* There is one shared device structure, so it can be accessed by the
 * ISR.  While this may raise concerns re: reentrancy, I currently
 * hold that since the device is read-only, nothing can get too messed
 * up. */

static struct e105_buttons_device_t buttons;

static void
e105_buttons_interrupt( int irq, void * dev_id, struct pt_regs * regs )
{
	__u8	new_status;

	new_status = *E105_BUTTONS;
	if( (buttons.status = new_status) != 0 ) {
		buttons.ready = 1;
		wake_up_interruptible( &buttons.waitq );
		if( buttons.fasync ) {
			kill_fasync( buttons.fasync, SIGIO );
		}
	}
}

static int fasync_e105_buttons( int fd, struct file * file_p, int on )
{
	return fasync_helper( fd, file_p, on, &buttons.fasync );
}

static int close_e105_buttons( struct inode * inode, struct file * file )
{
	fasync_e105_buttons( -1, file, 0 );
	if( --buttons.usage_count > 0 )
		return 0;
	
	free_irq( E105_BUTTONS_IRQ, E105_BUTTONS_DEV_ID );
	/* unsure if we have the right/need to touch GIUINTENL */
	buttons.ready = 0;

	MOD_DEC_USE_COUNT;
	return 0;
}

static int open_e105_buttons( struct inode * inode, struct file * file )
{
	if( buttons.usage_count++ > 0 )
		return 0;

	if( request_irq( E105_BUTTONS_IRQ, e105_buttons_interrupt,
			 SA_SHIRQ, "E105buttons", E105_BUTTONS_DEV_ID )) {
		--buttons.usage_count;
		return -EBUSY;
	}
	MOD_INC_USE_COUNT;
	*VR41XX_GIUINTTYPL   &= ~E105_BUTTONS_MASK;
	*VR41XX_GIUINTALSELL |=  E105_BUTTONS_MASK;
	*VR41XX_GIUINTENL    |=  E105_BUTTONS_MASK;
	*VR41XX_GIUINTSTATL  |=  E105_BUTTONS_MASK;
	return 0;
}

static ssize_t write_e105_buttons(struct file * file,
	const char * buffer, size_t count, loff_t *ppos)
{
	return -EINVAL;
}

static ssize_t read_e105_buttons(struct file * file,
	char * buffer, size_t count, loff_t *ppos)
{
	ssize_t 	rv;
	__u8		status;
	unsigned int	ready;
	
	if( count < 1 )
		return -EINVAL;

	if( (rv = verify_area( VERIFY_WRITE, buffer, count )) != 0 )
		return rv;
	
	while( 1 ) {
		/* read the button params atomically: */
		cli();
		ready = buttons.ready;
		status = buttons.status;
		buttons.status = 0;

		if( ready && status != 0 ) {
			sti();
			break;
		}

		/* either nothing to say or it's already been heard */
		if( file->f_flags & O_NONBLOCK ) {
			sti();
			return -EAGAIN;
		}

		/* keeping ints disabled ensures that the we don't call
		 * wake_up (from int handler) too soon when button press
		 * happens after we check buttons.ready, but before we
		 * put process to sleep see Rubini p209 */
		interruptible_sleep_on( &buttons.waitq );
		sti();

		/* this from Rubini p111 doesn't compile, probably
		 * because of the definition of sigset_t */
		/* if( current->signal & ~current->blocked ) */
		/* return -ERESTARTSYS; */

		/* let's try it this way */
		if( signal_pending( current ) )
			return -ERESTARTSYS;
	}

	__put_user( status, buffer );
	if( count > 1 )
		__clear_user( &buffer + 1, count - 1);

	return count;
}

static unsigned int poll_e105_buttons( struct file * file, poll_table * wait )
{
	poll_wait( file, &buttons.waitq, wait );
	return buttons.ready ? POLLIN | POLLRDNORM : 0;
}

static struct file_operations e105_buttons_fops = {
	NULL,
	read_e105_buttons,
	write_e105_buttons,
	NULL,
	poll_e105_buttons,
	NULL,
	NULL,
	open_e105_buttons,
	NULL,
	close_e105_buttons,
	NULL,
	fasync_e105_buttons
};

static struct miscdevice e105_buttons_device = {
	E105_BUTTONS_MINOR, "e105buttons", &e105_buttons_fops
};

int __init e105_buttons_init(void)
{
	buttons.status = 0;
	buttons.ready = 0;
	init_waitqueue_head( &buttons.waitq );
	buttons.usage_count = 0;
	buttons.fasync = NULL;
	misc_register( &e105_buttons_device );
// Shouldn't we check error condition here?
	return 0;
}

#ifdef MODULE
int init_module( void )
{
	return e105_buttons_init();
}

void cleanup_module( void )
{
	misc_deregister( &e105_buttons_device );
}
#endif
