/*
 * linux/arch/mips/vr41xx/prom.c
 *
 * Copyright (C) 1999 Bradley D. LaRonde and Michael Klar
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/init.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <asm/bootinfo.h>
#include <asm/addrspace.h>

char arcs_cmdline[CL_SIZE];
extern unsigned long mips_memory_upper;
extern int _end;

#define min(a,b)  (((a)<(b))?(a):(b))

unsigned long __init probe_ram_size(void)
{
	// this figures out the actual amount of RAM present, in increments of 1MB, max 64MB
	// returns the first physical address just past the end of RAM -
	// IOW the amount of ram detected

	unsigned long ramsize;
	unsigned int * p1, * p2;
	unsigned int d1, d2;

	// begin probe starting with the first 1MB past the global data area
	ramsize = ((unsigned long)&_end + 0x000fffff) & 0x1ff00000;

	// probe sequential 1MB areas, looking for a lack of RAM
	while (ramsize < 0x04000000) {
		// save val at a spot  and replace it with a pattern
		p1 = (unsigned int *)(KSEG1 + ramsize + 0x000ffff0);
		d1 = *p1;
		*p1 = 0x1234a5a5;
		barrier();

		// save a val at a second spot and replace with another pattern
		// this second write thwarts an empty bus faking us out
		p2 = (unsigned int *)(KSEG1 + ramsize + 0x000fffe0);
		d2 = *p2;
		*p2 = 0x00ff00ff;
		barrier();

		// isn't the first pattern still there?
		if (*p1 != 0x1234a5a5)
			// we are past the end of ram, so quit
			break;

		// restore what was there
		*p1 = d1;
		*p2 = d2;

		// check next 1MB area
		ramsize += 0x00100000;
	}

	// return the amount of ram detected
	return ramsize;
}

int __init prom_init(int argc, char **argv, char **envp)
{
	unsigned int mem_limit, mem_detected, mem_use;
	int i;

	// set upper limit to vr41xx maximum physical RAM (64MB)
	mem_limit = 64 << 20;

	// the bootloader passes us argc/argv[], but the kernel wants one big string
	// put it in arcs_cmdline, which later gets copied to command_line
	// (see arch/mips/kernel/setup.c)
	// we skip the first argument which is the program file name
	strcpy(arcs_cmdline, "");
	for (i = 1; i < argc; i++) {
		// is it the "mem=" option?
		if ( memcmp(argv[i], "mem=", 4) == 0 )
		{
			// limit to amount of memory specified in megabytes
			mem_limit = simple_strtoul(argv[i] + 4, 0, 10) << 20;

			printk("Command line argument limits memory to %dMB.\n", mem_limit >> 20);

			// don't pass this arg on to the kernel
			continue;
		}

		if ( strlen(arcs_cmdline) > 0 )
			strcat(arcs_cmdline, " ");
		strcat(arcs_cmdline, argv[i]);
	}

	mips_machgroup = MACH_GROUP_VR41XX;

	#ifdef CONFIG_VADEM_CLIO
	mips_machtype = MACH_VR41XX_VADEM_CLIO;
	#endif

	#ifdef CONFIG_CASIO_E105
	mips_machtype = MACH_VR41XX_CASIO_E105;
	#endif

	#ifdef CONFIG_EVEREX_FREESTYLE
	mips_machtype = MACH_VR41XX_EVEREX_FREESTYLE;
	#endif


	// probe for the amount of ram present
	// TODO: boot param "noramprobe" inhibits this probe; must use with boot param "ram"
	mem_detected = probe_ram_size();

	// don't use more RAM than we detected
	mem_use = min(mem_limit, mem_detected);

	// set the upper bound of usable memory
	mips_memory_upper = KSEG0 + mem_use;

	printk("Detected %dMB of memory.  Will use %dMB of it.\n",
		mem_detected >> 20, mem_use >> 20);

	return 0;
}

void __init prom_fixup_mem_map(unsigned long start, unsigned long end)
{
}

void prom_free_prom_memory(void)
{
}

