/*
 * linux/arch/mips/kernel/time.c
 *
 * Copyright (C) 1991, 1992, 1995  Linus Torvalds
 * Copyright (C) 1999 Bradley D. LaRonde
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>

#include <asm/bootinfo.h>
#include <asm/mipsregs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/vr41xx.h>

#include <linux/mc146818rtc.h>
#include <linux/timex.h>

int add_irq_action(int irq, struct irqaction *new);

extern volatile unsigned long lost_ticks;

#define USECS_PER_JIFFY (1000000/HZ)


#if 0

/* Cycle counter value at the previous timer interrupt.. */

static unsigned int timerhi = 0, timerlo = 0;

/*
 * On MIPS only R4000 and better have a cycle counter.
 *
 * FIXME: Does playing with the RP bit in c0_status interfere with this code?
 */
static unsigned long do_fast_gettimeoffset(void)
{
    u32 count;
    unsigned long res, tmp;

    /* Last jiffy when do_fast_gettimeoffset() was called. */
    static unsigned long last_jiffies = 0;
    unsigned long quotient;

    /*
     * Cached "1/(clocks per usec)*2^32" value.
     * It has to be recalculated once each jiffy.
     */
    static unsigned long cached_quotient = 0;

    tmp = jiffies;

    quotient = cached_quotient;

    if (last_jiffies != tmp) {
	last_jiffies = tmp;
	__asm__(".set\tnoreorder\n\t"
		".set\tnoat\n\t"
		".set\tmips3\n\t"
		"lwu\t%0,%2\n\t"
		"dsll32\t$1,%1,0\n\t"
		"or\t$1,$1,%0\n\t"
		"ddivu\t$0,$1,%3\n\t"
		"mflo\t$1\n\t"
		"dsll32\t%0,%4,0\n\t"
		"nop\n\t"
		"ddivu\t$0,%0,$1\n\t"
		"mflo\t%0\n\t"
		".set\tmips0\n\t"
		".set\tat\n\t"
		".set\treorder"
      :	"=&r"(quotient)
      :	"r"(timerhi),
		"m"(timerlo),
		"r"(tmp),
		"r"(USECS_PER_JIFFY)
      :	"$1");
	cached_quotient = quotient;
    }
    /* Get last timer tick in absolute kernel time */
    count = read_32bit_cp0_register(CP0_COUNT);

    /* .. relative to previous jiffy (32 bits is enough) */
    count -= timerlo;
//printk("count: %08lx, %08lx:%08lx\n", count, timerhi, timerlo);

    __asm__("multu\t%1,%2\n\t"
	    "mfhi\t%0"
  :	    "=r"(res)
  :	    "r"(count),
	    "r"(quotient));

    /*
     * Due to possible jiffies inconsistencies, we need to check 
     * the result so that we'll get a timer that is monotonic.
     */
    if (res >= USECS_PER_JIFFY)
	res = USECS_PER_JIFFY - 1;

    return res;
}

/* This function must be called with interrupts disabled 
 * It was inspired by Steve McCanne's microtime-i386 for BSD.  -- jrs
 * 
 * However, the pc-audio speaker driver changes the divisor so that
 * it gets interrupted rather more often - it loads 64 into the
 * counter rather than 11932! This has an adverse impact on
 * do_gettimeoffset() -- it stops working! What is also not
 * good is that the interval that our timer function gets called
 * is no longer 10.0002 ms, but 9.9767 ms. To get around this
 * would require using a different timing source. Maybe someone
 * could use the RTC - I know that this can interrupt at frequencies
 * ranging from 8192Hz to 2Hz. If I had the energy, I'd somehow fix
 * it so that at startup, the timer code in sched.c would select
 * using either the RTC or the 8253 timer. The decision would be
 * based on whether there was any other device around that needed
 * to trample on the 8253. I'd set up the RTC to interrupt at 1024 Hz,
 * and then do some jiggery to have a version of do_timer that 
 * advanced the clock by 1/1024 s. Every time that reached over 1/100
 * of a second, then do all the old code. If the time was kept correct
 * then do_gettimeoffset could just return 0 - there is no low order
 * divider that can be accessed.
 *
 * Ideally, you would be able to use the RTC for the speaker driver,
 * but it appears that the speaker driver really needs interrupt more
 * often than every 120 us or so.
 *
 * Anyway, this needs more thought....          pjsg (1993-08-28)
 * 
 * If you are really that interested, you should be reading
 * comp.protocols.time.ntp!
 */

#define TICK_SIZE tick

#endif


static unsigned long do_slow_gettimeoffset(void)
{
	// not implemented
	return 0;
}

static unsigned long (*do_gettimeoffset) (void) = do_slow_gettimeoffset;


// This version of gettimeofday has near microsecond resolution.
void do_gettimeofday(struct timeval *tv)
{
    unsigned long flags;

    save_and_cli(flags);
    *tv = xtime;
    tv->tv_usec += do_gettimeoffset();

    /*
     * xtime is atomically updated in timer_bh. lost_ticks is
     * nonzero if the timer bottom half hasnt executed yet.
     */
    if (lost_ticks)
	tv->tv_usec += USECS_PER_JIFFY;

    restore_flags(flags);

    if (tv->tv_usec >= 1000000) {
	tv->tv_usec -= 1000000;
	tv->tv_sec++;
    }
}

void do_settimeofday(struct timeval *tv)
{
    cli();
    /* This is revolting. We need to set the xtime.tv_usec
     * correctly. However, the value in this location is
     * is value at the last tick.
     * Discover what correction gettimeofday
     * would have done, and then undo it!
     */
    tv->tv_usec -= do_gettimeoffset();

    if (tv->tv_usec < 0) {
	tv->tv_usec += 1000000;
	tv->tv_sec--;
    }
    xtime = *tv;
    time_state = TIME_BAD;
    time_maxerror = MAXPHASE;
    time_esterror = MAXPHASE;
    sti();
}

/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick
 */
static void timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	// clear the interrupt
	*VR41XX_RTCINTREG = 0x2;

	do_timer(regs);

	// update the time from the RTC?
}

/* Converts Gregorian date to seconds since 1970-01-01 00:00:00.
 * Assumes input in normal date format, i.e. 1980-12-31 23:59:59
 * => year=1980, mon=12, day=31, hour=23, min=59, sec=59.
 *
 * [For the Julian calendar (which was used in Russia before 1917,
 * Britain & colonies before 1752, anywhere else before 1582,
 * and is still in use by some communities) leave out the
 * -year/100+year/400 terms, and add 10.]
 *
 * This algorithm was first published by Gauss (I think).
 *
 * WARNING: this function will overflow on 2106-02-07 06:28:16 on
 * machines were long is 32-bit! (However, as time_t is signed, we
 * will already get problems at other places on 2038-01-19 03:14:08)
 */
static inline unsigned long mktime(unsigned int year, unsigned int mon,
				   unsigned int day, unsigned int hour,
				   unsigned int min, unsigned int sec)
{
    if (0 >= (int) (mon -= 2)) {	/* 1..12 -> 11,12,1..10 */
	mon += 12;		/* Puts Feb last since it has leap day */
	year -= 1;
    }
    return (((
		 (unsigned long) (year / 4 - year / 100 + year / 400 + 367 * mon / 12 + day) +
		 year * 365 - 719499
	     ) * 24 + hour	/* now have hours */
	    ) * 60 + min	/* now have minutes */
	) * 60 + sec;		/* finally seconds */
}

struct irqaction timer_action = {
	timer_interrupt, SA_INTERRUPT, 0,
	"timer", NULL, NULL
};

void __init time_init(void)
{
	// xtime.tv_sec = mktime(year, mon, day, hour, min, sec);
	xtime.tv_sec = mktime(1967, 1, 3, 4, 45, 00);
	xtime.tv_usec = 0;

	// I chose RTCLong1 for the system timer
	// It has it's own cpu interrupt, is not T-Clock dependent, and has sufficient resolution

	// Set the RTCLong1 counter (32.768kHz) to expire in 1 / HZ second (HZ times per second)
	// Figure ticks/sec divided by jiffies/sec = ticks/jiffy
	// This rounds instead of truncates
	*VR41XX_RTCL1LREG = (32768 + HZ/2) / HZ;
	*VR41XX_RTCL1HREG = 0;

	// Add the timer action to the list for the RTCLong1 irq.
	// Put the action on the cpu interrupt, not the ICU interrupt.
	add_irq_action(VR41XX_IRQ_INT1, &timer_action);

	// Unmask the RTCLong1 interrupt in the ICU.
	// I unmask it here once and leave it unmasked.
	// The cpu interrupt is the one that gets masked/unmasked in do_IRQ.
	*VR41XX_MSYSINT1REG |= 0x4;
}

