/*
 *  linux/drivers/video/sed1345fb.c
 *
 *  Epson SED1354 Color Graphics LCD/CRT Controller
 * 
 *  Copyright (C) 1999 Bradley D. LaRonde.
 *
 *  Based on virtual frame buffer by Geert Uytterhoeven
 *  Based on simple frame buffer by Mike Klar
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <video/fbcon.h>
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb2.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <asm/addrspace.h>

#ifdef CONFIG_VADEM_CLIO
#include <asm/vr41xx.h>
const u_long videomemory = KSEG1 + VR41XX_LCD;
const u_long videomemorysize = 640 * 480;
#define FB_X_RES	640
#define FB_Y_RES	480
#define FB_BPP		8
#define FB_IS_GREY	0
#define FB_INVERSE	0
#endif

static int currcon = 0;
static struct display disp;
static struct fb_info my_fb_info;
static struct { u_char red, green, blue, pad; } palette[256];

static char my_fb_name[16] = "sed1354";

// fbcon uses 16 palette indicies
// support mapping through dispsw_data
static union {
	unsigned char cfb8[16];
} fbcon_cmap;

// helper functions

static u_long get_line_length(int xres_virtual, int bpp)
{
	u_long length;
	length = (xres_virtual * bpp + 7) >> 3;
	return(length);
}

static void get_fix_screeninfo(struct fb_fix_screeninfo *fix, struct fb_var_screeninfo *var)
{
	// return fixed display parameters
	memset(fix, 0, sizeof(struct fb_fix_screeninfo));
	strcpy(fix->id, my_fb_name);
	fix->smem_start = videomemory;
	fix->smem_len = videomemorysize;
	fix->type = FB_TYPE_PACKED_PIXELS;
	fix->type_aux = 0;
	switch (var->bits_per_pixel) {
	case 1:
	    fix->visual = FB_VISUAL_MONO01;
	    break;

	case 2:
	case 4:
	    fix->visual = FB_VISUAL_PSEUDOCOLOR;
	    break;

	// 8 bpp truecolor?  yup.  take a look at how the sed1354 does 8 bpp.
	// even though it goes though the LUTs, it only makes sense to use
	// a linear mapping and leave it alone - bdl
	case 8:
	case 16:
	    fix->visual = FB_VISUAL_TRUECOLOR;
	    break;
	}
	fix->ywrapstep = 1;
	fix->xpanstep = 1;
	fix->ypanstep = 1;
	fix->line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
}

static void get_color_bitfields(struct fb_var_screeninfo *var)
{
	switch (var->bits_per_pixel) {
	case 1:
	case 2:
	case 4:
		// pseudocolor
		var->red.offset = 0;
		var->red.length = var->bits_per_pixel;
		var->green.offset = 0;
		var->green.length = var->bits_per_pixel;
		var->blue.offset = 0;
		var->blue.length = var->bits_per_pixel;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;

	case 8:
		// RGB 332
		var->red.offset = 5;
		var->red.length = 3;
		var->green.offset = 2;
		var->green.length = 3;
		var->blue.offset = 0;
		var->blue.length = 2;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;

	case 16:
		// RGB 565
		var->red.offset = 0;
		var->red.length = 5;
		var->green.offset = 5;
		var->green.length = 6;
		var->blue.offset = 11;
		var->blue.length = 5;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	}

	var->red.msb_right = 0;
	var->green.msb_right = 0;
	var->blue.msb_right = 0;
	var->transp.msb_right = 0;
}

static int my_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                         u_int *transp, struct fb_info *info)
{
	// Read a single color register and split it into
	// colors/transparent. Return != 0 for invalid regno.

	if (regno > 255)
		return 1;
	*red = (palette[regno].red<<8) | palette[regno].red;
	*green = (palette[regno].green<<8) | palette[regno].green;
	*blue = (palette[regno].blue<<8) | palette[regno].blue;
	*transp = 0;

	return 0;
}

static int my_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                         u_int transp, struct fb_info *info)
{
	// Set a single color register. The values supplied are already
	// rounded down to the hardware's capabilities (according to the
	// entries in the var structure). Return != 0 for invalid regno.

	// So it says anyway.
	// From what I can tell, drivers/video/fbcmap.c passes in
	// un-rounded-down 16-bit values.
	
	if (regno > 255)
		return 1;

#ifdef FBCON_HAS_CFB8
	if( regno < 16 ) {
		// maintain index to rgb map for fbcon (accessed through dispsw_data)
		fbcon_cmap.cfb8[regno] =
			 ((red & 0xe000) >> 8)
			 | ((green & 0xe000) >> 11)
			 | ((blue & 0xc000) >> 14);
	}
#endif 

	red >>= 8;
	green >>= 8;
	blue >>= 8;

	palette[regno].red = red;
	palette[regno].green = green;
	palette[regno].blue = blue;

	return 0;
}

static void do_install_cmap(int con, struct fb_info *info)
{
	if (con != currcon)
		return;

	if (fb_display[con].cmap.len)
		fb_set_cmap(&fb_display[con].cmap, 1, my_setcolreg, info);
	else
		fb_set_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel), 1,
			my_setcolreg, info);
}


// framebuffer functions

static int my_fb_open(struct fb_info *info, int user)
{
	// Nothing, only a usage count for the moment
	MOD_INC_USE_COUNT;
	return(0);
}
        
static int my_fb_release(struct fb_info *info, int user)
{
	MOD_DEC_USE_COUNT;
	return(0);
}

static struct fb_var_screeninfo my_var_default = {
	FB_X_RES, FB_Y_RES, FB_X_RES, FB_Y_RES, 0, 0, FB_BPP, FB_IS_GREY,
	{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0},
	0, FB_ACTIVATE_NOW, -1, -1, 0, 20000, 64, 64, 32, 32, 64, 2,
	0, FB_VMODE_NONINTERLACED, {0,0,0,0,0,0}
};

static int my_fb_get_fix(struct fb_fix_screeninfo *fix, int con, struct fb_info *info)
{
	struct fb_var_screeninfo *var;

	if (con == -1)
		var = &my_var_default;
	else
		var = &fb_display[con].var;

	// fill in the fixed display params based on var
	get_fix_screeninfo(fix, var);

	return 0;
}

static int my_fb_get_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info)
{
	if (con == -1)
		*var = my_var_default;
	else
		*var = fb_display[con].var;

	// fill in the color bitfield info
	get_color_bitfields(var);

	return 0;
}

static int my_fb_set_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
	int err, activate = var->activate;
	int oldxres, oldyres, oldvxres, oldvyres, oldbpp;
	u_long line_length;

	struct display *display;
	if (con >= 0)
		display = &fb_display[con];
	else
		display = &disp; // used during initialization

	// FB_VMODE_CONUPDATE and FB_VMODE_SMOOTH_XPAN are equal!
	// as FB_VMODE_SMOOTH_XPAN is only used internally

	if (var->vmode & FB_VMODE_CONUPDATE) {
		var->vmode |= FB_VMODE_YWRAP;
		var->xoffset = display->var.xoffset;
		var->yoffset = display->var.yoffset;
	}

	// Some very basic checks
	if (!var->xres)
		var->xres = 1;
	if (!var->yres)
		var->yres = 1;
	if (var->xres > var->xres_virtual)
		var->xres_virtual = var->xres;
	if (var->yres > var->yres_virtual)
		var->yres_virtual = var->yres;
	if (var->bits_per_pixel <= 1)
		var->bits_per_pixel = 1;
	else if (var->bits_per_pixel <= 2)
		var->bits_per_pixel = 2;
	else if (var->bits_per_pixel <= 4)
		var->bits_per_pixel = 4;
	else if (var->bits_per_pixel <= 8)
		var->bits_per_pixel = 8;
	else if (var->bits_per_pixel <= 16)
		var->bits_per_pixel = 16;
	else
		return -EINVAL;

	// Memory limit
	line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
	if (line_length*var->yres_virtual > videomemorysize)
		return -ENOMEM;

	get_color_bitfields(var);

	if ((activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) {
		oldxres = display->var.xres;
		oldyres = display->var.yres;
		oldvxres = display->var.xres_virtual;
		oldvyres = display->var.yres_virtual;
		oldbpp = display->var.bits_per_pixel;
		display->var = *var;

		if (oldxres != var->xres || oldyres != var->yres ||
			oldvxres != var->xres_virtual || oldvyres != var->yres_virtual ||
			oldbpp != var->bits_per_pixel) {

			struct fb_fix_screeninfo fix;

			get_fix_screeninfo(&fix, var);

			display->screen_base = (char*)videomemory;
			display->visual = fix.visual;
			display->type = fix.type;
			display->type_aux = fix.type_aux;
			display->ypanstep = fix.ypanstep;
			display->ywrapstep = fix.ywrapstep;
			display->line_length = fix.line_length;
			display->can_soft_blank = 1;
			display->inverse = FB_INVERSE;

			switch (var->bits_per_pixel) {
#ifdef FBCON_HAS_MFB
			case 1:
				display->dispsw = &fbcon_mfb;
				break;
#endif

#ifdef FBCON_HAS_CFB2
			case 2:
				display->dispsw = &fbcon_cfb2;
				break;
#endif

#ifdef FBCON_HAS_CFB4
			case 4:
				display->dispsw = &fbcon_cfb4;
				break;
#endif

#ifdef FBCON_HAS_CFB8
			case 8:
				display->dispsw = &fbcon_cfb8;
				display->dispsw_data = fbcon_cmap.cfb8;
				break;
#endif

#ifdef FBCON_HAS_CFB16
			case 16:
				display->dispsw = &fbcon_cfb16;
				display->dispsw_data = fbcon_cmap.cfb16;
				break;
#endif
			default:
				display->dispsw = &fbcon_dummy;
				break;
			}

			if (my_fb_info.changevar)
				(*my_fb_info.changevar)(con);
		}

		if (oldbpp != var->bits_per_pixel) {
			if ((err = fb_alloc_cmap(&display->cmap, 0, 0)))
				return err;
			do_install_cmap(con, info);
		}
	}

	return 0;
}

static int my_fb_pan_display(struct fb_var_screeninfo *var, int con,
			   struct fb_info *info)
{
	// Pan or Wrap the Display
	// This call looks only at xoffset, yoffset and the FB_VMODE_YWRAP flag

	if (var->vmode & FB_VMODE_YWRAP) {
		if (var->yoffset < 0 ||
			var->yoffset >= fb_display[con].var.yres_virtual ||
			var->xoffset)
			return -EINVAL;
	}
	else {
		if (var->xoffset+fb_display[con].var.xres >
			fb_display[con].var.xres_virtual ||
			var->yoffset+fb_display[con].var.yres >
			fb_display[con].var.yres_virtual)
			return -EINVAL;
	}

	fb_display[con].var.xoffset = var->xoffset;
	fb_display[con].var.yoffset = var->yoffset;

	if (var->vmode & FB_VMODE_YWRAP)
		fb_display[con].var.vmode |= FB_VMODE_YWRAP;
	else
		fb_display[con].var.vmode &= ~FB_VMODE_YWRAP;

	return 0;
}

static int my_fb_get_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
	// Get the Colormap

	// current console?
	if (con == currcon)
		return fb_get_cmap(cmap, kspc, my_getcolreg, info);
	else if (fb_display[con].cmap.len) // non default colormap?
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
	else
		fb_copy_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel), cmap, kspc ? 0 : 2);

	return 0;
}

static int my_fb_set_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
	// Set the Colormap

	int err;

	// no colormap allocated?
	if (!fb_display[con].cmap.len) {
		if ((err = fb_alloc_cmap(&fb_display[con].cmap, 1<<fb_display[con].var.bits_per_pixel, 0)))
			return err;
	}

	// current console?
	if (con == currcon)
		return fb_set_cmap(cmap, kspc, my_setcolreg, info);
	else
		fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);

	return 0;
}

static int my_fb_ioctl(struct inode *inode, struct file *file, u_int cmd,
	u_long arg, int con, struct fb_info *info)
{
	return -EINVAL;
}


// console support

static int my_fbcon_switch(int con, struct fb_info *info)
{
	// Do we have to save the colormap?
	if (fb_display[currcon].cmap.len)
		fb_get_cmap(&fb_display[currcon].cmap, 1, my_getcolreg, info);

	currcon = con;
	// Install new colormap
	do_install_cmap(con, info);
	return 0;
}

static int my_fbcon_updatevar(int con, struct fb_info *info)
{
	// Update the `var' structure (called by fbcon.c)
	return 0;
}

static void my_fbcon_blank(int blank, struct fb_info *info)
{
	// Blank the display.
}

int __init sed1354_setup(char *options)
{
	char *this_opt;

	my_fb_info.fontname[0] = '\0';

	if (!options || !*options)
		return 0;

	for (this_opt = strtok(options, ","); this_opt; this_opt = strtok(NULL, ",")) {
		if (!strncmp(this_opt, "font:", 5))
			strcpy(my_fb_info.fontname, this_opt+5);
	}
	return 0;
}

static struct fb_ops my_fb_ops = {
	my_fb_open, my_fb_release, my_fb_get_fix, my_fb_get_var, my_fb_set_var, my_fb_get_cmap,
	my_fb_set_cmap, my_fb_pan_display, my_fb_ioctl
};

void __init set_recommended_lut_values(void)
{
	// these are the 8 bpp LUT values recommended by Epson
	unsigned char lutval[] = {
		0x00, 0x00, 0x00,
		0x03, 0x03, 0x05,
		0x05, 0x05, 0x0a,
		0x07, 0x07, 0x0f,
		0x09, 0x09, 0x00,
		0x0b, 0x0b, 0x00,
		0x0d, 0x0d, 0x00,
		0x0f, 0x0f, 0x00,
		0x00, 0x00, 0x00,
		0x00, 0x00, 0x00,
		0x00, 0x00, 0x00,
		0x00, 0x00, 0x00,
		0x00, 0x00, 0x00,
		0x00, 0x00, 0x00,
		0x00, 0x00, 0x00,
		0x00, 0x00, 0x00,
	};

	int i;

#ifdef CONFIG_VADEM_CLIO
	const unsigned int reg_base = videomemory + 0x200000;
#endif

	// set up to write lut entries auto-increment from 0
	*((volatile unsigned char*)reg_base + 0x24) = 0x00;

	barrier();

	// write lut entries
	for ( i = 0; i < sizeof(lutval); i++ )
		*((volatile unsigned char*)reg_base + 0x26) = lutval[i];

	// select bank 0 (only affects output)
	*((volatile unsigned char*)reg_base + 0x27) = 0x00;

	printk("LUTs set to recommended values.\n");
}

int __init sed1354_init(void)
{
	int result;

	strcpy(my_fb_info.modename, my_fb_name);
	my_fb_info.changevar = NULL;
	my_fb_info.node = -1;
	my_fb_info.fbops = &my_fb_ops;
	my_fb_info.disp = &disp;
	my_fb_info.switch_con = &my_fbcon_switch;
	my_fb_info.updatevar = &my_fbcon_updatevar;
	my_fb_info.blank = &my_fbcon_blank;
	my_fb_info.flags = FBINFO_FLAG_DEFAULT;

	my_fb_set_var(&my_var_default, -1, &my_fb_info);

	set_recommended_lut_values();

	result = register_framebuffer(&my_fb_info);
	if (result < 0)
	{
		printk(KERN_INFO "result %d registering frame buffer\n", result);
		return -EINVAL;
	}

	printk(KERN_INFO "fb%d: %s frame buffer device\n",
	       GET_FB_IDX(my_fb_info.node), my_fb_info.modename);

	printk(KERN_INFO "%d x %d, %d bpp, %ldK of video memory\n",
		my_var_default.xres, my_var_default.yres, my_var_default.bits_per_pixel,
		videomemorysize>>10);

	return 0;
}


#ifdef MODULE
int init_module(void)
{
	return sed1354_init();
}

void cleanup_module(void)
{
	unregister_framebuffer(&my_fb_info);
}
#endif /* MODULE */

